///////////////////////////////////////////////////////////////////////
// FileName   :  "infifo16.v"
// Author     :  Jordan (Jianping) Lu
// Department :  ECE of Concordia University
// Profile    :  This design is based on an industry documentation of ATM  
//               RCMP, and developed it out only for research purpose. 
//               For any questions and errors in this code, please contact
//               me by email: jianping@ece.concordia.ca or phone: 
//               (514)848-8758. 
////////////////////////////////////////////////////////////////////////
// This version has two modification with the original industry 
// documentation. One is that we only consider the situation of 16 bits 
// data path. We did not consider 8 bits of data bus case. This case can be 
// developed in a similar way. The other one is that we modify the polling 
// behavior in address model based on our own definition. 
// The design includes the following parts: FIFO process, FIFO control circuit 
// and  Polling process


`define  FWIDTH     16          // Width of the FIFO.
`define  FDEPTHIN   128         // Depth of the FIFO (contain 4 ATM cells).
`define  FCWIDTHIN  7           // Counter Width 
`define  DATWIDTH   16          // Width of the data cell
`define  PRTYWIDTH  2           // Width of the parity bits
`define  td         #2          // Register Q time


module IN_FIFO(  
                IFCLK,  
                IDAT,   
                ISOC,
                PRTY,   
                IAVALID_ICA4, 
                IADDR43,
                ICA32,
                ICA1,
                IADDR_IWRENB, 

                IWRENB1,
                IPOLL,
                HECUDF,
                ICAINV,
                CELLPOST,
                CELLLEN,
                IBYTEPRTY,  
                IPTYP,
                FIFODP,
                ICALEVEL0,
                IFRST,
                IFDAT,
                IFRDB,
                IFOCA,
                IFRCA,
                IPRTY,
                PHYCELL,
                FIFODATOUT,
                WRITEB,
                RD_PTR,
                WR_PTR
                );
                                        ///////////////////////////
                                        //Interface signal of RCMP
                                        //with PHY device
                                        ///////////////////////////

input                   IFCLK;          // FIFO clock  (52 MHz or lower)
input [(`DATWIDTH-1):0] IDAT;           // The input cell data 
input [(`PRTYWIDTH-1):0]PRTY;           // Parity bits
input                   ISOC;           // The input start of cell 
input                   IAVALID_ICA4;   // Address valid/ ICA[4]
input  [1:0]            ICA32;          
input                   IPOLL;          // Input polling select
input                   ICA1;           // ICA[1]
output [2:0]            IADDR_IWRENB;   // IADDR[2:0]/IWRENB[4:2]
output                  IWRENB1;
output [1:0]            IADDR43;


                                        // Since VIS does not support inout,
                                        // split inout to input and output,
                                        // respectively

                                        ///////////////////////////////
                                        // Internal signal for register 
                                        // 0x08 
                                        ///////////////////////////////

input                   HECUDF;         // HEC/UDF is included in cell?? (15)
input                   ICAINV;         // Select ICA[4:1] polarity, (14)
input [2:0]             CELLPOST;       // The # of the postpend, (13-10)
input [2:0]             CELLLEN;        // The # of the appended, (9-6)
input                   IBYTEPRTY;      // Byte parity or word parity ?? (5)
input                   IPTYP;          // Even or odd parity ?? (4)
output [1:0]            FIFODP;         // Cell depth (3-2)
input                   ICALEVEL0;      // ICA[4:1] interpretation(1)
input                   IFRST;          // Input FIFO reset (0)
output [1:0]            IPRTY;          // Parity error (Register 0x03)
output                  PHYCELL;        // Physical cell


                                        ///////////////////////////////
                                        // Internal signal for interconnect 
                                        // with uP, uBuf and output FIFO 
                                        ///////////////////////////////

input                    IFRDB;         // Read IN_FIFO ready
output [(`DATWIDTH-1):0] IFDAT;         // Data output of IN_FIFO
output                   IFRCA;         // Indication of IN_FIFO with one
                                        // cell space available
output                   IFOCA;         // Indication of IN_FIFO with at
                                        // least one cells
                                          
output                   WRITEB;        // The fifo interface signals
output                   RD_PTR;        
output                   WR_PTR;
                                          // Define output signals

reg [(`DATWIDTH-1):0]     IFDAT; 
reg                       IFOCA;        
reg                       IFRCA; 
reg                       ICA_S;
wire [2:0]                IADDR_IWRENB;
reg  [2:0]                IADDR_IWRENB_S;
  
                                       // Define internal signals

reg    [`FCWIDTHIN-1:0]     IFCOUNTER;   // Counter indicates num of data
                                         // in FIFO
reg    [(`FCWIDTHIN-1):0]   RD_PTR;      // Current read pointer of 16 bus
reg    [(`FCWIDTHIN-1):0]   WR_PTR;      // Current write pointer of16 bus 
reg    [5:0]              CELLCOUNTER; // Counter in each cell write
reg    [4:0]              IADDR;       // Address in ipoll=1 model
reg    [4:0]              IADDR_S;     // Sample of address
reg                       ICA_S1;
reg    [1:0]              p_state;     // Present state
reg                       WRITEB;
reg    [1:0]              IPRTY;
reg    [23:0]             VPI_VCI;     // Store the VPI/VCI bits
reg                       CLP;         // Store the Cell lose priority bit
reg                       DISCARD;     // Discard cell 
reg                       PHYCELL;

wire                      PRTYCHKL;
wire                      PRTYCHKH;
wire                      ICA1i;       //Intermediate signal of ICA1
input   [(`FWIDTH-1):0]    FIFODATOUT;  //Data out from the FIFO Mem
wire   [2:0]              realCELLLEN; //Real cell length
wire   [2:0]              realCELLPOST;//Real postpend cell length

wire   [3:0]              CELLPRE=realCELLLEN-realCELLPOST;
                                        //Prepend cell length 

wire   [(`FWIDTH-1):0]    FIFODATIN = IDAT;
reg 			  p_addr_3;
reg [4:0] 		  addr_r1;
reg [4:0]		  addr_r2;
reg [4:0]		  addr_r3;
reg [1:0] 		  p_state_r1;
reg [1:0]		  p_state_r2;
reg [1:0] 		  p_state_r3;
reg			  pause_r1;
reg			  pause_r2;

// Initialization
initial begin
          p_state_r1 = 0;
          IADDR = 0;
          addr_r1 = 0;
          CLP = 0;
          addr_r2 = 0;
          pause_r2 = 0;
          IFDAT = 0;
          IADDR_IWRENB_S = 0;
          addr_r3 = 0;
          p_state = 0;
          DISCARD = 0;
          IFRCA = 0;
          WRITEB = 0;
          pause_r1 = 0;
          ICA_S = 0;
          ICA_S1 = 0;
          CELLCOUNTER = 0;
          PHYCELL = 0;
          RD_PTR = 0;
          IFOCA = 0;
          VPI_VCI = 0;
          IFCOUNTER = 0;
          IADDR_S = 0;
          p_state_r3 = 0;
          IPRTY = 0;
          WR_PTR = 0;
          p_state_r2 = 0;
          p_addr_3 = 0;
     end
          
         
//Store the VPI/VCI and CLP bits for each cell. If VPI_VCI bits are all 
//zero, then the cell is unassigned cell or physical layer OAM cell. If 
//further CLP=0, it is unassigned cell; PHYCELL is zero. If CLP =1, it 
//is PHYSIC Layer OAM cell, PHYCELL is one. In both case, the cell is discarded.
//DISCARD is "1"


always @(posedge IFCLK or IFRST) 

begin

        if (IFRST) begin

                VPI_VCI = 24'hffffff;
                CLP = 1'b0;

        end

        else if (CELLCOUNTER==CELLPRE)

                VPI_VCI[23:12]= IDAT[11:0];

        else if (CELLCOUNTER==CELLPRE+1)  begin

                VPI_VCI[11:0] = IDAT[15:4];
                CLP = IDAT[0];

        end

        else begin

                VPI_VCI = VPI_VCI;
                CLP = CLP;
        end

end


always @(posedge IFCLK or IFRST) 

begin 

	if (IFRST == 1) 
	
		PHYCELL = 0;
        else if (CELLCOUNTER==CELLPRE+2 && VPI_VCI==24'h0 && ~CLP)

                PHYCELL = 1;

        else 
                PHYCELL = 0;



end


// ISOC should happen with the first word of a cell, but it may not be happen
// in  each cell. If it happen in the other word of a cell, this cell will be
// discarded

always @(posedge IFCLK or IFRST) 

begin 

        if (IFRST == 1)
        
          DISCARD = 0;
          
	else if ((ISOC && CELLCOUNTER != 1) || (CELLCOUNTER==CELLPRE+2 
			&& VPI_VCI==24'h0)) 

                DISCARD = 1;

        else if (CELLCOUNTER == 0)
        
        	DISCARD = 0;
        
end




// Parity checking is done in IDAT bus. IF the parity checking is failed, the 
// information will be stored in Register0x03 (IPRTY[1:0]). IPTYP is used 
// for definition of parity type. If set to logic 1, input IPRTY[1] is the 
// even parity selector bit IDAT[15:8]; and if IPRTY[0] is the even parity 
// selector for IDAT[7:0]. When IPTYP is logic 0, it is for odd parity 
// checking. 

// IBYTEPRTY define word parity checking or byte parity checking. If IBYTEPRTY
// is logic 1; Byte parity checking. If IBYTEPRTY is 0, word parity checking.
// only PRTY[1] works and PRTY[0] is ignored.



assign  PRTYCHKL=IDAT[0]^IDAT[1]^IDAT[2]^IDAT[3]
                ^IDAT[4]^IDAT[5]^IDAT[6]^IDAT[7];

assign  PRTYCHKH=IDAT[9]^IDAT[10]^IDAT[11]^IDAT[12]
                ^IDAT[13]^IDAT[14]^IDAT[15]^IDAT[8];



always @(posedge IFCLK or IFRST) 

begin 

	if (IFRST)
	
		IPRTY = 0;
		
        else if (IBYTEPRTY && IPTYP) begin

                IPRTY[0]= PRTYCHKL^PRTY[0];
                IPRTY[1]= PRTYCHKH^PRTY[1];

        end

        else if (IBYTEPRTY && ~IPTYP) begin

                IPRTY[0]=~(PRTYCHKL^PRTY[0]);
                IPRTY[1]=~(PRTYCHKH^PRTY[1]);

        end

        else if (~IBYTEPRTY && IPTYP) begin 

                IPRTY[1]=PRTYCHKL^PRTYCHKH^PRTY[1];
                IPRTY[0]=1'bx;

        end

        else begin

                IPRTY[1]=~(PRTYCHKL^PRTYCHKH^PRTY[1]);
                IPRTY[0]=1'bx;

        end

end
  


// Output assignment
// Purpose: assign output to uP registers

always @(posedge IFCLK) 

begin

        if (IFRST)


          IFDAT = 16'hfffe ; //fffe denotes high impedence


        else if (~IFRDB)  

                IFDAT = FIFODATOUT[(`DATWIDTH-1):0] ;
 

        else  

                IFDAT = 16'hfffe;

 


end 

// Purpose: get an instance of FIFO_BLK
/*    
  FIFO_BLK_IN    memblk(.clk(IFCLK),
                        .writeN(WRITEB),
                        .rd_addr(RD_PTR),
                        .wr_addr(WR_PTR),
                        .data_in(FIFODATIN),
                        .data_out(FIFODATOUT)
                       ); 

*/

// Control circuitry for FIFO. If IFRST signal is asserted
// all the counters are set to 0. IF write only the write counter
// is incremented else if read only read counter is incremented
// else if both, read and write counters are incremented.
// IFCOUNTER indicates the num of items in the fifo. Write only
// increments  the IFCOUNTER, read only decrements the counter and
// read && write doen't change the counter value. If HECUDF is set to 
// logic 0, the third word of 27-word ATM cell.
    
    
always @(posedge IFCLK or IFRST)

begin

          if(IFRST) 

                    WR_PTR = 0;
                    
          else if ((CELLCOUNTER == 27 + CELLLEN) && HECUDF && DISCARD) 
          
          	WR_PTR = WR_PTR - 27 - CELLLEN;
                    
          else if ((CELLCOUNTER == 26 + CELLLEN) && ~HECUDF && DISCARD) 
          
          	WR_PTR = WR_PTR - 26 - CELLLEN;
    
end

always @(posedge IFCLK or IFRST) 

begin 


        if (IFRST) 

                IFCOUNTER = 0;
                
	else if ((CELLCOUNTER == 27 + CELLLEN) && HECUDF && DISCARD) 
          
        
        begin

                if (~IFRDB)

                 IFCOUNTER = IFCOUNTER-(28+CELLLEN);

                else 

                 IFCOUNTER = IFCOUNTER-(27+CELLLEN);

        end

 	else if ((CELLCOUNTER == 26 + CELLLEN) && ~HECUDF && DISCARD) 

        begin 

                if (~IFRDB)

                 IFCOUNTER = IFCOUNTER-(27+CELLLEN);

                else 

                 IFCOUNTER = IFCOUNTER-(26+CELLLEN);

        end 


        else if (~WRITEB && IFRDB  && ((CELLCOUNTER !=2+CELLPRE && ~HECUDF)
                || HECUDF)) 
 
                  IFCOUNTER = IFCOUNTER + 1;

        else if (WRITEB && ~IFRDB  && ((CELLCOUNTER!=2+CELLPRE &&               
                ~HECUDF) || HECUDF)) 

                  IFCOUNTER = IFCOUNTER - 1;

        else 

                  IFCOUNTER = IFCOUNTER;

end


always @(posedge IFCLK or IFRST) 

begin 
             
               
        if (IFRST) 

                 RD_PTR = 0;

        else if (~IFRDB)
      
                 RD_PTR = RD_PTR + 1;
 
        else 

                 RD_PTR = RD_PTR;
           
               
end



// IFOCA indicate there is a cell available or not. If there is an cell, IFRDB can be
// asserted, otherwise IFRDB is deasserted. If IFOCA assert, IFRDB should be deasserted in 
// the cycle for some time before IFRDB is asserted. So in this case IFOCA expresses 
// the exactly at least one cell.
   
always @(posedge IFCLK or  IFRST )

begin

     if(IFRST)
   
                 IFOCA = 1'b0;


     else if ((((IFCOUNTER==27+CELLLEN && IFRDB==1'b0 && WRITEB==1'b1)||
                (IFCOUNTER==26+CELLLEN && (!(IFRDB==1'b1 && WRITEB==1'b0))) 
                || IFCOUNTER<=25+CELLLEN) && HECUDF ) || 
	     (((IFCOUNTER==26+CELLLEN && IFRDB==1'b0 && WRITEB==1'b1)||
                (IFCOUNTER==25+CELLLEN && (!(IFRDB==1'b1 && WRITEB==1'b0))) 
                || IFCOUNTER<=24+CELLLEN) && ~HECUDF ))
 
                 IFOCA = 1'b0;


      else 

                 IFOCA = 1'b1;
      
end


       

// IFRCA indicates that the FIFO has at least one cell space. when the input device 
// justify write is not available. So IFRCA means that exactly at least once cell free.

always @(posedge IFCLK or IFRST )

begin
 
     if(IFRST)
 
                 IFRCA = 1'b1;
 
     else if ((((IFCOUNTER==(`FDEPTHIN-28-CELLLEN) && WRITEB==1'b0 && IFRDB==1'b1 )
                || (IFCOUNTER==(`FDEPTHIN-27-CELLLEN) && (!(WRITEB==1'b1 && IFRDB==1'b0)))
                || IFCOUNTER >=(`FDEPTHIN-26-CELLLEN)) && ~HECUDF) ||
  		(((IFCOUNTER==(`FDEPTHIN-28-CELLLEN) && WRITEB==1'b0 && IFRDB==1'b1 )
                || (IFCOUNTER==(`FDEPTHIN-27-CELLLEN) && (!(WRITEB==1'b1 && IFRDB==1'b0)))
                || IFCOUNTER >=(`FDEPTHIN-26-CELLLEN)) && ~HECUDF))              
       
                 IFRCA = 1'b0;
 
    
     else 

                 IFRCA = 1'b1;
    
     
end


// CELLLEN and CELLPOST provide the information of the cell length      
// and header position. By these two signals, WR_PTR IFCOUNTER,                 
// HECUDF and IFDEPTHIN functions can be decided.
// FIFODP(0): 4CELLS, FIFODP(1): 3CELLS,FIFODP(2): 2CELLS,
// FIFODP(3): 1CELL; 
// CELLLEN: MAX 5 for 16 bits bus;
// CELLPOST: MAX 5 for 16 bits bus, and it should       
// also be less than CELLLEN, but will not check in this model.   
// when HECUDF=0, 26 words for 16 bits bus


assign  realCELLLEN=(CELLLEN > 5) ? 5 : CELLLEN;

assign  realCELLPOST=(CELLPOST > 5) ? 5 : CELLPOST;

assign  FIFODP =(HECUDF) ? ((IFCOUNTER<(27+CELLLEN))   ? 2'b11 :
                        (IFCOUNTER<((27+CELLLEN)+(27+CELLLEN))) ? 2'b10 :
                        (IFCOUNTER<((27+CELLLEN)+(27+CELLLEN)+
                                    (27+CELLLEN)))
                         ? 2'b01 : 2'b00) :
                        ((IFCOUNTER<(26+CELLLEN))   ? 2'b11 :
                        (IFCOUNTER<((26+CELLLEN)+(26+CELLLEN))) ? 2'b10 :
                        (IFCOUNTER<((26+CELLLEN)+(26+CELLLEN)+(26+CELLLEN)))
                         ? 2'b01 : 2'b00);
                        
        

//Cell counter process 
//Purpose: it is used to control polling and FIFO process

always @(posedge IFCLK or IFRST) 

begin 

        if (IFRST) 

                 CELLCOUNTER = 0;

        else if (((CELLCOUNTER ==26 + CELLLEN) && ~HECUDF) || 
        	((CELLCOUNTER ==27 + CELLLEN) && HECUDF))
        	
                 CELLCOUNTER = 0;

        else if (~WRITEB) 

                 CELLCOUNTER = CELLCOUNTER + 1;

        else 

                 CELLCOUNTER = CELLCOUNTER;

end


//Polling process
//Purpose: define WRITEB value in the different polling cases
//Describe the relations among IAVALID_ICA4, IADDR, ICA32, ICA1, IADDR_IWRENB 
//and IWRENB1 when IPOLL is 1 or 0. When IPOLL=0, FIFO receive ICA, and 
//generate IWRENB signals through round-robin algorithm; when IPOLL=1, FIFO 
//gives IADDR signals and receives ICA1, and then generate IWRENB1


//Assign value to IWRENB1, IADDR_IWRENB;

assign IWRENB1= IPOLL ? WRITEB : 
        (p_state == 2'b00) ? WRITEB : 1;

assign IADDR_IWRENB[0] = IPOLL ? IADDR[0] : 
                (p_state==2'b01) ? WRITEB : 1;

assign IADDR_IWRENB[1] = IPOLL ? IADDR[1] : 
                (p_state==2'b10) ? WRITEB : 1;

assign IADDR_IWRENB[2] = IPOLL ? IADDR[2] : 
                (p_state==2'b11) ? WRITEB : 1;

// define WRITEB signal. WRITEB will be asserted to '0' when p_state and valid 
// IADDR keep the same value for at least three clock cycle when CELLCOUNTER is 
// the maximum value (27+ celllen or 26+celllen depended on HECUDF) or 0.

always @(posedge IFCLK or IFRST) begin 

	if (IFRST) WRITEB = 1;
	
	else if (((CELLCOUNTER == 26+CELLLEN) && HECUDF) ||
		((CELLCOUNTER == 25+CELLLEN) && ~HECUDF))
		
		WRITEB = 1;
		
	else if (p_addr_3 == 1 && CELLCOUNTER == 0)
		
		WRITEB = 0;
		
end 

// define p_addr_3 signal, this signal means that P_state or addr keep in the 
// same value for three clock cycle when CELLCOUNTER is equal to the maximum
// value or 0. 

always @ (posedge IFCLK or IFRST) begin

	if (IFRST)
          begin
		p_addr_3 = 0;
		addr_r1 = 0;
		addr_r2 = 0;
		addr_r3 = 0;
          end
	else if (IPOLL && ((CELLCOUNTER==27+CELLLEN && HECUDF) || 
		(CELLCOUNTER==26+CELLLEN && ~HECUDF) || CELLCOUNTER == 0))
               begin 
		addr_r1 = IADDR;
		addr_r2 = addr_r1;
		addr_r3 = addr_r2;
		if (addr_r1 == addr_r2 && addr_r2==addr_r3 
		&& CELLCOUNTER == 0 && addr_r1 != 0)
			p_addr_3 = 1;
		else if (addr_r1 == 0 && addr_r2 == 0 && addr_r3 == 0 
			&& CELLCOUNTER == 0)
                       begin
			pause_r1 = 1;
			pause_r2 = pause_r1;
                        end
		 	if (pause_r1 == 1 && pause_r2 == 1 && addr_r1 == 0
		 	&& addr_r2 == 0 && addr_r3 == 0)
                          begin
                            p_addr_3 = 1;
                            pause_r1 = 0;
                            pause_r2 = 0;
                          end
                end
	else if (~IPOLL && ((CELLCOUNTER==27+CELLLEN && HECUDF) || 
		(CELLCOUNTER==26+CELLLEN && ~HECUDF) || CELLCOUNTER == 0))
                begin
		p_state_r1 = p_state;
		p_state_r2 = p_state_r1;
		p_state_r3 = p_state_r2;
		if (p_state_r1 == p_state_r2 && p_state_r2==p_state_r3 
		&& CELLCOUNTER == 0 && p_state_r1 != 0)
			p_addr_3 = 1;
		else if (p_state_r1 == 0 && p_state_r2 == 0 && p_state_r3 == 0 
			&& CELLCOUNTER == 0)
                       begin
			pause_r1 = 1;
			pause_r2 = pause_r1;
                       end
		 	if (pause_r1 == 1 && pause_r2 == 1 && p_state_r1 == 0
		 	&& p_state_r2 == 0 && p_state_r3 == 0)
                          begin
                            p_addr_3 = 1;
                            pause_r1 = 0;
                            pause_r2 = 0;
                          end
                 end
	else p_addr_3 = 0;
end		
		
// define the signal p_state, p_state will seach 

always @(posedge IFCLK or IFRST) begin 

	if (IFRST)
		p_state = 0;
		
	else if (~IPOLL && ((CELLCOUNTER==27+CELLLEN && HECUDF) || 
		(CELLCOUNTER==26+CELLLEN && ~HECUDF) || CELLCOUNTER == 0)) 
		
		case (p_state) 
		
		2'b00: if (ICA32[0] == 1)
			p_state = 2'b01;
		       else if (ICA32[1] == 1)
		        p_state = 2'b10;
		       else if (IAVALID_ICA4 == 1)
		        p_state = 2'b11;
		       else if (ICA1 == 1)
		        p_state = 2'b00;
		       else p_state = 2'b01;
		2'b01: if (ICA32[1] == 1)
		        p_state = 2'b10;
		       else if (IAVALID_ICA4 == 1)
		        p_state = 2'b11;
		       else if (ICA1 == 1)
		        p_state = 2'b00;
		       else if (ICA32[0] == 1)
		        p_state = 2'b01;
		       else p_state = 2'b10;
		2'b10:  if (IAVALID_ICA4 == 1)
		        p_state = 2'b11;
		       else if (ICA1 == 1)
		        p_state = 2'b00;
		       else if (ICA32[0] == 1)
			p_state = 2'b01;
		       else if (ICA32[1] == 1)
		        p_state = 2'b10;
		       else p_state = 2'b11;
		2'b11: if (ICA1 == 1)
		        p_state = 2'b00;
		       else if (ICA32[0] == 1)
			p_state = 2'b01;
		       else if (ICA32[1] == 1)
		        p_state = 2'b10;
		       else if (IAVALID_ICA4 == 1)
		        p_state = 2'b11;
		       else p_state = 2'b00;
	      endcase	
 end
assign IADDR43 = IPOLL ? IADDR[1:0] : 2'b11;


assign ICA1i = IPOLL ? (ICAINV ? ~ICA1 : ICA1) : 1'b1;


always @(posedge IFCLK or IFRST)  begin 

        if (IFRST) 

        begin

                WRITEB = 1;
                IADDR_IWRENB_S = 3'b111;
                p_state = 2'b00;
                IADDR = 5'b00000;


        end

        else if  (~IPOLL) begin 

        begin: ICA_S_assign

        if (~ICAINV)

                ICA_S = (p_state==2'b00) ? ICA1 :
                        (p_state==2'b01) ? ICA32[0] :
                        (p_state==2'b10) ? ICA32[1] : IAVALID_ICA4;

        else 

                ICA_S = (p_state==2'b00) ? ~ICA1 :
                        (p_state==2'b01) ? ~ICA32[0] :
                        (p_state==2'b10) ? ~ICA32[1] : ~IAVALID_ICA4;

        end


        begin 
        
        if (CELLCOUNTER==26+CELLLEN) 

                     WRITEB = 1;

        else if (CELLCOUNTER==27+CELLLEN) 

        begin


                     WRITEB = 1;
                        p_state = p_state+1;

        end

        else   if ((CELLCOUNTER==0 && IFRCA && ICA_S) || CELLCOUNTER!=0) 


                     WRITEB = 0;

        else 

                     WRITEB = 1;


        end

        end


        else 

        begin 

        begin

                if (!ICA1i) 

                begin 

                     IADDR = IADDR + 1;
                        ICA_S = 0;

                end 

                else if (CELLCOUNTER!=0) 

                begin 
                        
                     ICA_S1 = 1;
                        IADDR = IADDR;
                        IADDR_S = IADDR -1;
                end 

                else 

                begin 

                     ICA_S1 = 1;
                        IADDR = IADDR + 1;
                        IADDR_S = IADDR_S;

                end 
        end

        begin

                if (CELLCOUNTER==0 && ICA_S1) 

                     ICA_S = 1;

                else 

                     ICA_S = 0;

        end


        begin

                if (CELLCOUNTER==26+CELLLEN || CELLCOUNTER==27+CELLLEN) 

                     WRITEB = 1;             

                else if ((CELLCOUNTER ==0 && IFRCA && ICA_S)
                         || !(CELLCOUNTER==0 ||
                        CELLCOUNTER==26+CELLLEN || CELLCOUNTER==27+CELLLEN)) 
 
                     WRITEB = 0;
        
                else 

                     WRITEB = 1;


        end
       end

end
                         

endmodule
/*

///////////////////////////////////////////////////////////////////
// Configurable memory block for fifo. The width of the mem
// block is configured via "FWIDTH". All the data into fifo is done 
// synchronous to block.
//
// Author : Jordan (Jianping) Lu
///////////////////////////////////////////////////////////////////

module FIFO_BLK_IN(
                 clk,
                 writeN,
                 wr_addr,
                 rd_addr,
                 data_in,
                 data_out
                );


input                    clk;       // input clk.
input                    writeN;    // Write Signal to put datainto fifo.
input  [(`FCWIDTHIN-1):0]  wr_addr;   // Write Address.
input  [(`FCWIDTHIN-1):0]  rd_addr;   // Read Address.
input  [(`FWIDTH-1):0]   data_in;   // DataIn in to Memory Block
output [(`FWIDTH-1):0]   data_out;  // Data Out from the Memory Block(FIFO)

wire   [(`FWIDTH-1):0]   data_out;  
reg    [(`FWIDTH-1):0]   FIFO[(`FDEPTHIN-1):0];


assign data_out = FIFO[rd_addr];


always @(posedge clk)

begin

        if(writeN==1'b0)

                FIFO[wr_addr] = data_in;

end

endmodule


*/
